from dataclasses import dataclass, field
from typing import Union, TypeAlias, TypeVar

from typing_extensions import Self

from yupy.icomparable_schema import ComparableSchema, EqualityComparableSchema
from yupy.ischema import _SchemaExpectedType
from yupy.locale import locale
from yupy.validation_error import ErrorMessage, ValidationError, Constraint

__all__ = ('NumberSchema',)

_T = TypeVar('_T')
_NumberType: TypeAlias = Union[int, float]


@dataclass
class NumberSchema(ComparableSchema, EqualityComparableSchema):
    _type: _SchemaExpectedType = field(init=False, default=(float, int))

    def positive(self, message: ErrorMessage = locale["positive"]) -> Self:
        return self.gt(0, message)

    def negative(self, message: ErrorMessage = locale["negative"]) -> Self:
        return self.lt(0, message)

    def integer(self, message: ErrorMessage = locale["integer"]) -> Self:
        def _(x: _NumberType) -> None:
            if (x % 1) != 0:
                raise ValidationError(Constraint("integer", message), invalid_value=x)

        return self.test(_)

    # def truncate(self):
    #     ...

    # def round(self, method: Literal['ceil', 'floor', 'round', 'trunc']) -> 'NumberSchema':
    #     self._transforms

    def multiple_of(self, multiplier: Union[int, float],
                    message: ErrorMessage = locale["multiple_of"]) -> Self:
        def _(x: Union[int, float]) -> None:
            if x % multiplier != 0:
                raise ValidationError(Constraint("multiple_of", message, multiplier), invalid_value=x)

        return self.test(_)



from typing import Protocol, Any, runtime_checkable

from typing_extensions import Self

from yupy.locale import locale
from yupy.schema import Schema
from yupy.validation_error import ErrorMessage, ValidationError, Constraint

__all__ = (
    'IEqualityComparableSchema',
    'IComparableSchema',
    'EqualityComparableSchema',
    'ComparableSchema',
)


@runtime_checkable
class IEqualityComparableSchema(Protocol):
    def eq(self, value: Any, message: ErrorMessage = locale["eq"]) -> Self: ...

    def ne(self, value: Any, message: ErrorMessage = locale["ne"]) -> Self: ...


@runtime_checkable
class IComparableSchema(Protocol):
    def le(self, limit: Any, message: ErrorMessage = locale["le"]) -> Self: ...

    def ge(self, limit: Any, message: ErrorMessage = locale["ge"]) -> Self: ...

    def lt(self, limit: Any, message: ErrorMessage = locale["lt"]) -> Self: ...

    def gt(self, limit: Any, message: ErrorMessage = locale["gt"]) -> Self: ...


class EqualityComparableSchema(Schema):
    def eq(self, value: Any, message: ErrorMessage = locale["eq"]) -> Self:
        def _(x: Any) -> None:
            if x != value:
                raise ValidationError(Constraint("eq", message, value), invalid_value=x)

        return self.test(_)

    def ne(self, value: Any, message: ErrorMessage = locale["ne"]) -> Self:
        def _(x: Any) -> None:
            if x == value:
                raise ValidationError(Constraint("ne", message, value), invalid_value=x)

        return self.test(_)


class ComparableSchema(Schema):

    def le(self, limit: Any, message: ErrorMessage = locale["le"]) -> Self:
        def _(x: Any) -> None:
            if x > limit:
                raise ValidationError(Constraint("le", message, limit), invalid_value=x)

        return self.test(_)

    def ge(self, limit: Any, message: ErrorMessage = locale["ge"]) -> Self:
        def _(x: Any) -> None:
            if x < limit:
                raise ValidationError(Constraint("ge", message, limit), invalid_value=x)

        return self.test(_)

    def lt(self, limit: Any, message: ErrorMessage = locale["lt"]) -> Self:
        def _(x: Any) -> None:
            if x >= limit:
                raise ValidationError(Constraint("lt", message, limit), invalid_value=x)

        return self.test(_)

    def gt(self, limit: Any, message: ErrorMessage = locale["gt"]) -> Self:
        def _(x: Any) -> None:
            if x <= limit:
                raise ValidationError(Constraint("gt", message, limit), invalid_value=x)

        return self.test(_)



from dataclasses import field, dataclass
from typing import Any, List, Optional

from typing_extensions import Self

from yupy.ischema import TransformFunc, ValidatorFunc, _SchemaExpectedType
from yupy.locale import locale
from yupy.validation_error import ErrorMessage, ValidationError, Constraint

__all__ = ('Schema',)


@dataclass
class Schema:  # Implement ISchema
    _type: _SchemaExpectedType = field(default=object)
    _transforms: List[TransformFunc] = field(init=False, default_factory=list)
    _validators: List[ValidatorFunc] = field(init=False, default_factory=list)
    _optional: bool = True
    _required: ErrorMessage = locale["required"]
    _nullability: bool = False
    _not_nullable: ErrorMessage = locale["not_nullable"]
    _default: Optional[_SchemaExpectedType] = None

    @property
    def optional(self) -> bool:
        return self._optional

    @property
    def nullability(self) -> bool:
        return self._nullability

    def required(self, message: ErrorMessage = locale["required"]) -> Self:
        self._required: ErrorMessage = message
        self._optional: bool = False
        return self

    def not_required(self) -> Self:
        self._optional: bool = True
        return self

    def nullable(self) -> Self:
        self._nullability: bool = True
        return self

    def not_nullable(self, message: ErrorMessage = locale["not_nullable"]) -> Self:
        self._nullability: bool = False
        self._not_nullable: ErrorMessage = message
        return self

    def _nullable_check(self, value: Any) -> None:
        if not self._nullability:
            raise ValidationError(
                Constraint("nullable", self._not_nullable),
                invalid_value=value
            )

    def _type_check(self, value: Any) -> None:
        type_ = self._type
        if type_ is Any:
            return
        if not isinstance(value, type_):
            raise ValidationError(
                Constraint("type", locale["type"], type_, type(value)),
                invalid_value=value
            )

    def transform(self, func: TransformFunc) -> Self:
        self._transforms: List[TransformFunc]
        self._transforms.append(func)
        return self

    def test(self, func: ValidatorFunc) -> Self:
        self._validators.append(func)
        return self

    def _transform(self, value: Any) -> Any:
        transformed: Any = value
        for t in self._transforms:
            transformed = t(transformed)
        return transformed

    def default(self, value: Optional[_SchemaExpectedType]) -> Self:
        self._default = value
        return self

    def const(self, value: Optional[_SchemaExpectedType], message: ErrorMessage = locale["const"]) -> Self:
        def _(x: Any) -> None:
            if x != value:
                raise ValidationError(Constraint("const", message, value), invalid_value=x)
        return self.test(_)

    def validate(self, value: Any, abort_early: bool = True, path: str = "~") -> Any:
        try:
            if value is None:
                self._nullable_check(value)
                if self._default is not None:
                    value = self._default
                else:
                    return None

            transformed = self._transform(value)

            self._type_check(transformed)

            for v in self._validators:
                v(transformed)
            return transformed
        except ValidationError as err:
            raise ValidationError(err.constraint, path, invalid_value=value)



from typing import Optional, Literal, TypedDict

from yupy.validation_error import ErrorMessage

__all__ = (
    'locale',
    'set_locale'
)


class Locale(TypedDict, total=False):
    const: ErrorMessage
    type: ErrorMessage
    min: ErrorMessage
    max: ErrorMessage
    length: ErrorMessage
    required: ErrorMessage
    not_nullable: ErrorMessage
    test: ErrorMessage
    matches: ErrorMessage
    email: ErrorMessage
    url: ErrorMessage
    uuid: ErrorMessage
    lowercase: ErrorMessage
    uppercase: ErrorMessage
    le: ErrorMessage
    ge: ErrorMessage
    lt: ErrorMessage
    gt: ErrorMessage
    eq: ErrorMessage
    ne: ErrorMessage
    integer: ErrorMessage
    multiple_of: ErrorMessage
    positive: ErrorMessage
    negative: ErrorMessage
    array_of: ErrorMessage
    shape: ErrorMessage
    shape_values: ErrorMessage
    one_of: ErrorMessage
    undefined: ErrorMessage


LocaleKey = Literal[
    "const", "type", "min", "max", "length", "required", "not_nullable", "test", "matches",
    "email", "url", "uuid", "lowercase", "uppercase",
    "le", "ge", "lt", "gt", "eq", "ne",
    "integer", "multiple_of", "positive", "negative", "array_of", "shape",
    "shape_values", "one_of", "undefined"
]

locale: Locale = {
    "const": lambda args: "Value is not match the const %r" % args,
    "type": lambda args: "Value is not of type %r, got %r" % args,
    "min": lambda args: "Min length must be %r" % args,
    "max": lambda args: "Max length must be %r" % args,
    "length": lambda args: "Length must be %r" % args,
    "uppercase": "Value must be an uppercase string",
    "lowercase": "Value must be a lowercase string",
    "required": "Field is required",
    "not_nullable": "Value can't be null",
    "test": "Test failed",
    'matches': "Don't match regex",  # FIXME
    "email": "Value must be a valid email",
    "url": "Value must be a valid URL",
    "uuid": "Value must be a valid UUID",
    "le": lambda args: "Value must be less or equal to %r" % args,
    "ge": lambda args: "Value must be greater or equal to %r" % args,
    "lt": lambda args: "Value must be less than %r" % args,
    "gt": lambda args: "Value must be greater than %r" % args,
    "eq": lambda args: "Value must be equal to %r" % args,
    "ne": lambda args: "Value must be not equal to %r" % args,
    "positive": "Value must be positive, a.g. > 0",
    "negative": "Value must be positive, a.g. < 0",
    "integer": "Value must be valid 'int', got 'float'",
    "array_of": lambda args: "Schema must be a type of Schema, got %r" % args,
    "multiple_of": lambda args: "Value must be a multiple of %r" % args,
    "shape": "'shape' must be a type of 'Shape'",
    "shape_values": "all shape items must have a values of type of Schema",
    "one_of": lambda args: "Must be one of %r" % args,
    "undefined": "undefined validation error"
}


def set_locale(locale_: Optional[Locale] = None) -> Locale:
    if locale_:
        locale.update(locale_)
    return locale


def get_error_message(key: LocaleKey) -> ErrorMessage:
    return locale.get(key, "undefined")



from dataclasses import dataclass, field
from typing import Any, List, Iterable

from typing_extensions import Self

from yupy.icomparable_schema import ComparableSchema, EqualityComparableSchema
from yupy.ischema import _SchemaExpectedType, ISchema
from yupy.isized_schema import SizedSchema
from yupy.locale import locale
from yupy.schema import Schema
from yupy.util.concat_path import concat_path
from yupy.validation_error import ErrorMessage, Constraint, ValidationError

__all__ = ('ArraySchema',)


@dataclass
class ArraySchema(SizedSchema, ComparableSchema, EqualityComparableSchema):
    _type: _SchemaExpectedType = field(init=False, default=(list, tuple))
    _fields: List[ISchema] = field(init=False, default_factory=list)
    _type_of: Schema = field(init=False, default_factory=Schema)

    # @property
    # def fields(self) -> List[ISchema]:
    #     return self._fields

    def of(self, schema: ISchema, message: ErrorMessage = locale["array_of"]) -> Self:
        if not isinstance(schema, Schema):
            raise ValidationError(Constraint("array_of", message, type(schema)), invalid_value=schema)
        self._type_of = schema
        return self

    def validate(self, value: Any, abort_early: bool = True, path: str = "~") -> Any:
        super().validate(value, abort_early, path)
        return self._validate_array(value, abort_early, path)  # Convert tuple to list for iteration

    def _validate_array(self, value: Iterable, abort_early: bool = True,
                        path: str = "~") -> Iterable:
        errs: List[ValidationError] = []
        for i, v in enumerate(value):
            path_ = concat_path(path, i)
            try:
                self._type_of.validate(v, abort_early, path_)
            except ValidationError as err:
                if abort_early:
                    raise ValidationError(err.constraint, path_, invalid_value=v)
                else:
                    errs.append(err)
        if errs:
            raise ValidationError(
                Constraint('array', 'invalid array', path),
                path, errs, invalid_value=value)
        return value

    def __getitem__(self, item: int) -> ISchema:
        return self._fields[item]



from dataclasses import dataclass, field
from typing import Any, Iterable

from typing_extensions import Self

from yupy.ischema import _SchemaExpectedType
from yupy.locale import locale
from yupy.schema import Schema
from yupy.validation_error import ErrorMessage, Constraint, ValidationError

__all__ = ('MixedSchema',)


@dataclass
class MixedSchema(Schema):
    _type: _SchemaExpectedType = field(default=object)

    def one_of(self, items: Iterable, message: ErrorMessage = locale['one_of']) -> Self:
        """
        Adds a validation to check if the value is one of the provided items.
        """

        def _(x: Any) -> None:
            if x not in items:
                raise ValidationError(Constraint('one_of', message, items), invalid_value=x)

        return self.test(_)



from types import UnionType
from typing import Protocol, TypeVar, Callable, Any, TypeAlias, runtime_checkable, List, Type, Optional, Union

from typing_extensions import Self

from yupy.validation_error import ErrorMessage

_T = TypeVar("_T")
_SchemaExpectedType: TypeAlias = Union[type[_T], UnionType, tuple[Any, ...]]
_P = TypeVar('_P', bound=_SchemaExpectedType)

TransformFunc: TypeAlias = Callable[[Any], Any]
ValidatorFunc: TypeAlias = Callable[[_T], None]


@runtime_checkable
class ISchema(Protocol[_P]):
    _type: Type[_P]
    _transforms: List[TransformFunc]
    _validators: List[ValidatorFunc]
    _optional: bool
    _required: ErrorMessage
    _nullability: bool
    _not_nullable: ErrorMessage
    _default: Optional[_P] = None

    @property
    def optional(self) -> bool: ...

    @property
    def nullability(self) -> bool: ...

    def required(self, message: ErrorMessage) -> Self: ...

    def not_required(self) -> Self: ...

    def nullable(self) -> Self: ...

    def not_nullable(self, message: ErrorMessage) -> Self: ...

    def transform(self, func: TransformFunc) -> Self: ...

    def _transform(self, value: Any) -> Any: ...

    def test(self, func: ValidatorFunc) -> Self: ...

    def validate(self, value: Any, abort_early: bool = True, path: str = "") -> _P: ...

    def default(self, value: Optional[_P]) -> Self: ...

    def const(self, value: Optional[_P], message: ErrorMessage) -> Self: ...

    def _nullable_check(self, value: Any) -> None: ...

    def _type_check(self, value: Any) -> None: ...



from .array_schema import *
from .icomparable_schema import IComparableSchema, ComparableSchema
from .ischema import ISchema
from .isized_schema import ISizedSchema, SizedSchema
from .locale import *
from .mapping_schema import *
from .mixed_schema import *
from .number_schema import *
from .schema import *
from .string_schema import *
from .validation_error import *

string = StringSchema
number = NumberSchema
mapping = MappingSchema
array = ArraySchema
mixed = MixedSchema

__all__ = (
    'ValidationError',
    'Constraint',

    'Schema',
    'StringSchema',
    'NumberSchema',
    'MappingSchema',
    'ArraySchema',
    'MixedSchema',

    'ISchema',
    'IComparableSchema',
    'ComparableSchema',
    'ISizedSchema',
    'SizedSchema',

    'string',
    'number',
    'mapping',
    'array',
    'mixed',

    'locale',
    'set_locale',

    'util',
)



from typing import Protocol, Any, runtime_checkable

__all__ = ('Comparable',)


@runtime_checkable
class Comparable(Protocol):
    """
    A protocol for types that support comparison operations.  Provides
    default implementations for all rich comparison methods
    based on __eq__ and __lt__.
    """

    def __eq__(self, other: Any) -> bool:
        """
        Equality comparison.
        """
        raise NotImplementedError

    def __lt__(self, other: Any) -> bool:
        """
        Less-than comparison.
        """
        raise NotImplementedError

    def __le__(self, other: Any) -> bool:
        """
        Less-than-or-equal-to comparison.
        """
        return (self < other) or (self == other)  # Default implementation

    def __gt__(self, other: Any) -> bool:
        """
        Greater-than comparison.
        """
        return not (self <= other)  # Default implementation

    def __ge__(self, other: Any) -> bool:
        """
        Greater-than-or-equal-to comparison.
        """
        return not (self < other)  # Default implementation

    def __ne__(self, other: Any) -> bool:
        """
        Not-equal-to comparison.
        """
        return not (self == other)  # Default implementation



from dataclasses import dataclass, field
from typing import Any, MutableMapping, TypeAlias
from typing_extensions import Self

from yupy.icomparable_schema import EqualityComparableSchema
from yupy.ischema import _SchemaExpectedType, ISchema
from yupy.locale import locale
from yupy.schema import Schema
from yupy.util.concat_path import concat_path
from yupy.validation_error import ValidationError, Constraint

__all__ = ('MappingSchema',)

_SchemaShape: TypeAlias = MutableMapping[str, ISchema[Any]]


@dataclass
class MappingSchema(EqualityComparableSchema):
    _type: _SchemaExpectedType = field(init=False, default=dict)
    _fields: _SchemaShape = field(init=False, default_factory=dict)

    # @property
    # def fields(self) -> _SchemaShape:
    #     return self._fields

    def shape(self, fields: _SchemaShape) -> Self:
        if not isinstance(fields, dict):  # Перевірка залишається на dict, оскільки shape визначається через dict
            raise ValidationError(
                Constraint("shape", locale["shape"])
            )
        for key, item in fields.items():
            if not isinstance(item, ISchema):
                raise ValidationError(
                    Constraint("shape_values", locale["shape_values"]),
                    key,
                    invalid_value=item
                )
        self._fields = fields
        return self

    def validate(self, value: Any, abort_early: bool = True, path: str = "~") -> Any:
        super().validate(value, abort_early, path)
        return self._validate_shape(value, abort_early, path)

    def _validate_shape(self, value: MutableMapping[str, Any],
                        abort_early: bool = True,
                        path: str = "~") -> MutableMapping[str, Any]:

        errs: list[ValidationError] = []
        for k, f in self._fields.items():
            path_ = concat_path(path, k)
            try:
                if k not in value:
                    if not self._fields[k]._optional:
                        raise ValidationError(
                            Constraint("required", self._fields[k]._required, path_),
                            path_, invalid_value=value
                        )

                    value[k] = self._fields[k].validate(None, abort_early, path_)
                else:
                    value[k] = self._fields[k].validate(value[k], abort_early, path_)
            except ValidationError as err:
                if abort_early:
                    raise ValidationError(err.constraint, path_, invalid_value=value)
                errs.append(err)
        if errs:
            raise ValidationError(
                Constraint('object', 'invalid object', path),
                path, errs, invalid_value=value
            )
        return value

    def __getitem__(self, item: str) -> ISchema:
        return self._fields[item]



import re
from dataclasses import field, dataclass

from typing_extensions import Self

from yupy.icomparable_schema import ComparableSchema, EqualityComparableSchema
from yupy.ischema import _SchemaExpectedType
from yupy.isized_schema import SizedSchema
from yupy.locale import locale
from yupy.validation_error import ErrorMessage, ValidationError, Constraint

__all__ = ('StringSchema',)

rUUID_pattern = re.compile(
    r"^(?:[0-9a-f]{8}-[0-9a-f]{4}-[1-5][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}|00000000-0000-0000-0000-000000000000)$",
    re.IGNORECASE,
)

rEmail_pattern = re.compile(
    r"^[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$",
    re.IGNORECASE,
)

rUrl_pattern = re.compile(
    r"^((https?|ftp):)?//(((([a-z]|\d|-|\.|_|~|[ -퟿豈-﷏ﷰ-￯])|(%[\da-f]{2})|[!$&'()*+,;=]|:)*@)?(((\d|[1-9]\d|1\d\d|2[0-4]\d|25[0-5])\.(\d|[1-9]\d|1\d\d|2[0-4]\d|25[0-5])\.(\d|[1-9]\d|1\d\d|2[0-4]\d|25[0-5])\.(\d|[1-9]\d|1\d\d|2[0-4]\d|25[0-5]))|((([a-z]|\d|[ -퟿豈-﷏ﷰ-￯])|(([a-z]|\d|[ -퟿豈-﷏ﷰ-￯])([a-z]|\d|-|\.|_|~|[ -퟿豈-﷏ﷰ-￯])*([a-z]|\d|[ -퟿豈-﷏ﷰ-￯])))\.)+(([a-z]|[ -퟿豈-﷏ﷰ-￯])|(([a-z]|[ -퟿豈-﷏ﷰ-￯])([a-z]|\d|-|\.|_|~|[ -퟿豈-﷏ﷰ-￯])*([a-z]|[ -퟿豈-﷏ﷰ-￯])))\.?)(:\d*)?)(/((([a-z]|\d|-|\.|_|~|[ -퟿豈-﷏ﷰ-￯])|(%[\da-f]{2})|[!$&'()*+,;=]|:|@)+(/(([a-z]|\d|-|\.|_|~|[ -퟿豈-﷏ﷰ-￯])|(%[\da-f]{2})|[!$&'()*+,;=]|:|@)*)*)?)?(\?((([a-z]|\d|-|\.|_|~|[ -퟿豈-﷏ﷰ-￯])|(%[\da-f]{2})|[!$&'()*+,;=]|:|@)|[-]|/|\?)*)?(#((([a-z]|\d|-|\.|_|~|[ -퟿豈-﷏ﷰ-￯])|(%[\da-f]{2})|[!$&'()*+,;=]|:|@)|/|\?)*)?$",
    re.IGNORECASE,
)


@dataclass
class StringSchema(SizedSchema, ComparableSchema, EqualityComparableSchema):
    _type: _SchemaExpectedType = field(init=False, default=str)

    # def matches(self, regex: re.Pattern, message: ErrorMessage, exclude_empty: bool = False) -> Schema:
    #     def _(x: str):
    #         if not re.match(regex, x):
    #             raise ValidationError(message)
    #     self._validators.append(_)
    #     return self

    def email(self, message: ErrorMessage = locale["email"]) -> Self:
        def _(x: str) -> None:
            if not re.match(rEmail_pattern, x):
                raise ValidationError(Constraint("email", message), invalid_value=x)

        return self.test(_)

    def url(self, message: ErrorMessage = locale["url"]) -> Self:
        def _(x: str) -> None:
            if not re.match(rUrl_pattern, x):
                raise ValidationError(Constraint("url", message), invalid_value=x)

        return self.test(_)

    def uuid(self, message: ErrorMessage = locale["uuid"]) -> Self:
        def _(x: str) -> None:
            if not re.match(rUUID_pattern, x):
                raise ValidationError(Constraint("uuid", message), invalid_value=x)

        return self.test(_)

    # def datetime(self, message: ErrorMessage, precision: int, allow_offset: bool = False):
    #     def _(x: str):
    #         if ...:
    #             raise ValidationError(message)
    #     self._validators.append(_)
    #     return self

    def ensure(self) -> Self:
        def _(x: str) -> str:
            return x if x else ""

        self._transforms.append(_)
        return self

    # def trim(self, message: ErrorMessage):
    #     ...

    def lowercase(self, message: ErrorMessage = locale["lowercase"]) -> Self:
        def _(x: str) -> None:
            if x.lower() != x:
                raise ValidationError(Constraint("lowercase", message), invalid_value=x)

        return self.test(_)

    def uppercase(self, message: ErrorMessage = locale["uppercase"]) -> Self:
        def _(x: str) -> None:
            if x.upper() != x:
                raise ValidationError(Constraint("uppercase", message), invalid_value=x)

        return self.test(_)



from typing import Union

__all__ = ('concat_path',)


def concat_path(path: str, item: Union[str, int]) -> str:
    """
    Concatenates a path string with an item (either a string or an integer)
    to create a new path.

    Args:
        path: The current path string.
        item: The item to append to the path. If it's a string, it's joined
              with a dot; if it's an integer, it's appended in square brackets.

    Returns:
        The concatenated path string.

    Raises:
        TypeError: If the item is not a string or an integer.
    """
    if isinstance(item, int):
        item = f"[{item!r}]"
    if isinstance(item, str):
        if not path:
            return item
        return "/".join((path, item))
    else:
        raise TypeError("Unsupported item type")



from dataclasses import dataclass
from typing import Sized, Protocol, runtime_checkable

from typing_extensions import Self

from yupy.locale import locale
from yupy.schema import Schema
from yupy.validation_error import ErrorMessage, ValidationError, Constraint

__all__ = ('ISizedSchema', 'SizedSchema')


@runtime_checkable
class ISizedSchema(Protocol):

    def length(self, limit: int, message: ErrorMessage = locale["length"]) -> Self: ...

    def min(self, limit: int, message: ErrorMessage = locale["min"]) -> Self: ...

    def max(self, limit: int, message: ErrorMessage = locale["max"]) -> Self: ...


@dataclass
class SizedSchema(Schema):

    def length(self, limit: int, message: ErrorMessage = locale["length"]) -> Self:
        def _(x: Sized) -> None:  # Use Sized instead of Iterable
            if len(x) != limit:
                raise ValidationError(Constraint("length", message, limit), invalid_value=x)

        return self.test(_)

    def min(self, limit: int, message: ErrorMessage = locale["min"]) -> Self:
        def _(x: Sized) -> None:  # Use Sized instead of Iterable
            if len(x) < limit:
                raise ValidationError(Constraint("min", message, limit), invalid_value=x)

        return self.test(_)

    def max(self, limit: int, message: ErrorMessage = locale["max"]) -> Self:
        def _(x: Sized) -> None:  # Use Sized instead of Iterable
            if len(x) > limit:
                raise ValidationError(Constraint("max", message, limit), invalid_value=x)

        return self.test(_)



from dataclasses import dataclass, field
from typing import Generator, Any, List, Optional, Union, Callable, TypeAlias

__all__ = (
    'ErrorMessage',
    'ValidationError',
    'Constraint',
)

from typing_extensions import Self

ErrorMessage: TypeAlias = Union[str, Callable[[Any | List[Any]], str]]


@dataclass
class Constraint:
    type: str
    args: Any
    message: ErrorMessage = field(repr=False)

    def __init__(self,
                 type: Optional[str],
                 message: Optional[ErrorMessage] = None,
                 *args: Any,
                 ):
        self.type = type or "undefined"
        self.args = args
        if not message:
            from yupy.locale import get_error_message
            self.message = get_error_message("undefined")
        else:
            self.message = message

    @property
    def format_message(self) -> str:
        if callable(self.message):
            return self.message(self.args)
        return self.message


class ValidationError(ValueError):
    def __init__(
            self, constraint: Optional[Constraint] = None, path: str = "",
            errors: Optional[List['ValidationError']] = None,
            invalid_value: Optional[Any] = None, *args) -> None:
        if not constraint:
            self.constraint = Constraint("undefined")
        else:
            self.constraint = constraint
        self.path = path
        self._errors: List[ValidationError] = errors or []
        self.invalid_value: Optional[Any] = invalid_value
        super().__init__(self.path, self.constraint, self._errors, *args)

    def __str__(self) -> str:
        return "(path=%r, constraint=%r, message=%r)" % (self.path, self.constraint, self.constraint.format_message)

    def __repr__(self) -> str:
        return "ValidationError%s" % self.__str__()

    @property
    def errors(self) -> Generator['ValidationError', None, None]:
        yield self
        for error in self._errors:
            yield from error.errors

    @property
    def message(self) -> str:
        return "%r:%s" % (self.path, self.constraint.format_message)

    @property
    def messages(self) -> Generator[Union[property, str], None, None]:
        for e in self.errors:
            yield e.message
